import logging
import pathlib
from typing import Optional

import jaxtyping
import tiktoken
import torch
from typeguard import typechecked as typechecker

from toyllm.device import current_device
from toyllm.model.config import GPTModelConfig, gpt_config_124_m
from toyllm.model.gpt import GPTModel
from toyllm.tokenizer import gpt2_tokenizer, text_to_token_ids, token_ids_to_text

logger = logging.getLogger(__name__)


class TextGenerator:
    def __init__(
        self,
        model_config: Optional[GPTModelConfig] = None,
        model_instance: Optional[GPTModel] = None,
        model_file_path: Optional[pathlib.Path] = None,
        tokenizer: tiktoken.Encoding = gpt2_tokenizer,
        seed: int = 42,
    ):
        self.model_config = model_config
        self.model_instance = model_instance
        self.model_file_path = model_file_path
        self.tokenizer = tokenizer
        self.seed = seed

        self.gpt_model = self.__get_gpt_model()

    def __get_gpt_model(self) -> GPTModel:
        torch.manual_seed(self.seed)
        if self.model_instance is not None:
            model = self.model_instance
            self.model_config = model.config
        elif self.model_config is not None:
            model = GPTModel(self.model_config)
            # TODO: load mode weight
            if self.model_file_path is not None:
                model.load_state_dict(torch.load(self.model_file_path))
            else:
                logger.warning("Debug mode: with random model weight")
        else:
            raise ValueError("Can not initialize GPT Model without model_instance or model_config")
        # disable dropout and so on
        model.eval()
        if model.device != current_device:
            model.to(current_device)
        return model

    @property
    def context_length(self) -> int:
        assert self.model_config is not None, "Model config is None"
        return self.model_config.ctx_len

    def generate(
        self,
        prompt_text: str,
        max_gen_tokens: int = 10,
        top_k: Optional[int] = None,
        temperature: Optional[float] = None,
    ) -> str:
        """

        :param prompt_text: prompt text
        :param max_gen_tokens: maximum number of tokens to generate
        :param top_k: only keep `top_k`(logits) candidate tokens to select from.
        A little `top_k` will reduce the randomness of generated output.
        `top_k` must be greater than 0, like 5, 10 and so on.
        :param temperature: "Temperatures greater than 1 will result in more uniformly distributed token probabilities
        after applying the softmax; temperatures smaller than 1 will result in
        more confident (sharper or more peaky) distributions after applying the softmax"
        (https://github.com/rasbt/LLMs-from-scratch/blob/main/ch05/01_main-chapter-code/ch05.ipynb)
        The default temperature value is 0.6 in llama2.

        """
        # prompt text to tokens: (1, n_tokens)
        prompt_tokens = text_to_token_ids(prompt_text, self.tokenizer).to(self.gpt_model.device)

        for _ in range(max_gen_tokens):
            # Crop current context if it exceeds the supported context size(ctx_len)
            # E.g., if LLM supports only 5 tokens, and the context size is 10
            # then only the last 5 tokens are used as context

            # (batch, n_tokens) --(crop context)--> (batch, n_tokens' = min(ctx_len, n_tokens))
            context_text_token_ids = prompt_tokens[:, -self.context_length :]

            # Get the predictions
            # use `inference_mode` rather than `no_grad`(https://stackoverflow.com/questions/74191070)
            with torch.inference_mode():
                # (batch, n_token') --(forward)--> (batch, n_token', vocab_size)
                logits = self.gpt_model(context_text_token_ids)

            # Focus only on the last time step
            # (batch, n_tokens', vocab_size) --(keep last time step token)--> (batch, vocab_size)
            logits = logits[:, -1, :]

            # logits filter & scale
            if top_k is not None:
                logits = self._logits_top_k_filter(logits, top_k)
            if temperature is not None:
                probs = self._logits_temperature_scale(logits, temperature)
                # Sample from the scaled multinomial distribution
                # (batch, vocab_size)--(keep the max prob token)--> (batch, 1)
                next_token_id = torch.multinomial(probs, num_samples=1)
            else:
                # Get the idx of the vocab entry with the highest logits value
                # (batch, vocab_size)--(keep the max prob token)--> (batch, 1)
                next_token_id = torch.argmax(logits, dim=-1, keepdim=True)

            # Append sampled index to the running sequence
            # (batch, n_tokens') --(append next token)--> (batch, n_tokens' + 1)
            prompt_tokens = torch.cat((prompt_tokens, next_token_id), dim=1)

        generate_text = token_ids_to_text(prompt_tokens)

        return generate_text

    @jaxtyping.jaxtyped(typechecker=typechecker)
    @staticmethod
    def _logits_top_k_filter(
        logits: jaxtyping.Float[torch.Tensor, "batch_size vocab_size"],
        top_k: int,
    ) -> jaxtyping.Float[torch.Tensor, "batch_size vocab_size"]:
        """
        ref1: https://github.com/rasbt/LLMs-from-scratch/blob/62fb11d5e0449a6d49bda7337d6cfa5a735718da/ch05/01_main-chapter-code/generate.py#L166-L185
        ref2: https://github.com/huggingface/transformers/blob/c4d4e8bdbd25d9463d41de6398940329c89b7fb6/src/transformers/generation_utils.py#L903-L941
        ref3: https://github.com/meta-llama/llama/blob/main/llama/generation.py#L188-L192
        """
        top_k = min(top_k, logits.size(-1))  # make sure top_k <= vocab size
        top_k_logits, _top_k_indexes = torch.topk(logits, k=top_k, dim=-1)
        min_logit = top_k_logits[:, -1]
        logits = torch.where(logits < min_logit, torch.tensor(float("-inf")).to(logits.device), logits)
        return logits

    @jaxtyping.jaxtyped(typechecker=typechecker)
    @staticmethod
    def _logits_temperature_scale(
        logits: jaxtyping.Float[torch.Tensor, "batch_size vocab_size"],
        temperature: float,
    ) -> jaxtyping.Float[torch.Tensor, "batch_size vocab_size"]:
        logits = logits / temperature
        probs = torch.softmax(logits, dim=-1)
        return probs


if __name__ == "__main__":
    text_generator = TextGenerator(model_config=gpt_config_124_m)

    prompt_text = "Hello, I am"
    generate_text = text_generator.generate(prompt_text=prompt_text, top_k=10, temperature=0.9)
    print(generate_text)



# Copyright 2024 Xiangzhuang Shen
# Copyright 2023-2024 Sebastian Raschka
# SPDX-License-Identifier: Apache-2.0


from typing import TypeAlias

import jaxtyping
import torch
import torch.nn as nn
from typeguard import typechecked as typechecker

from toyllm.model.config import GPTModelConfig

GPTInputType: TypeAlias = jaxtyping.Int[torch.Tensor, "batch_size num_tokens"]
GPTInnerType: TypeAlias = jaxtyping.Float[torch.Tensor, "batch_size num_tokens emb_dim"]
GPTOutputType: TypeAlias = jaxtyping.Float[torch.Tensor, "batch_size num_tokens vocab_size"]


class MultiHeadAttention(nn.Module):
    def __init__(
        self,
        d_in: int,
        d_out: int,
        ctx_len: int,
        dropout_rate: float,
        n_heads: int,
        qkv_bias: bool = False,
    ):
        super().__init__()
        assert d_out % n_heads == 0, "d_out must be divisible by num_heads"

        self.d_out = d_out
        self.n_heads = n_heads
        self.head_dim = d_out // n_heads  # Reduce the projection dim to match desired output dim

        self.Wq = nn.Linear(d_in, d_out, bias=qkv_bias)  # Query Weight
        self.Wk = nn.Linear(d_in, d_out, bias=qkv_bias)  # Key Weight
        self.Wv = nn.Linear(d_in, d_out, bias=qkv_bias)  # Value Weight
        self.out_proj = nn.Linear(d_out, d_out)  # Linear layer to combine head outputs
        self.dropout = nn.Dropout(dropout_rate)
        self.register_buffer("mask", torch.triu(torch.ones(ctx_len, ctx_len), diagonal=1))

    @jaxtyping.jaxtyped(typechecker=typechecker)
    def forward(self, x: GPTInnerType) -> GPTInnerType:
        batch_size, num_tokens, _d_in = x.shape

        # (batch_size, num_tokens, d_in) -> (batch_size, num_tokens, d_out)
        keys = self.Wk(x)
        queries = self.Wq(x)
        values = self.Wv(x)

        # We implicitly split the matrix by adding a `num_heads` dimension
        # Unroll last dim: (batch_size, num_tokens, d_out) -> (batch_size, num_tokens, num_heads, head_dim)
        keys = keys.view(batch_size, num_tokens, self.n_heads, self.head_dim)
        values = values.view(batch_size, num_tokens, self.n_heads, self.head_dim)
        queries = queries.view(batch_size, num_tokens, self.n_heads, self.head_dim)

        # Transpose: (batch_size, num_tokens, num_heads, head_dim) -> (batch_size, num_heads, num_tokens, head_dim)
        keys = keys.transpose(1, 2)
        queries = queries.transpose(1, 2)
        values = values.transpose(1, 2)

        # Compute scaled dot-product attention (aka self-attention) with a causal mask
        # attn_scores shape: (batch_size, num_heads, num_tokens, num_tokens)
        attn_scores = queries @ keys.transpose(2, 3)  # Dot product for each head

        # Original mask truncated to the number of tokens and converted to boolean
        mask_bool = self.mask.bool()[:num_tokens, :num_tokens]

        # Use the mask to fill attention scores
        attn_scores.masked_fill_(mask_bool, -torch.inf)

        attn_weights = torch.softmax(attn_scores / keys.shape[-1] ** 0.5, dim=-1)
        # TODO: explain why dropout here
        attn_weights = self.dropout(attn_weights)

        # Shape: (b, num_tokens, num_heads, head_dim)
        context_vec = (attn_weights @ values).transpose(1, 2)

        # Combine heads, where self.d_out = self.num_heads * self.head_dim
        context_vec = context_vec.contiguous().view(batch_size, num_tokens, self.d_out)
        context_vec = self.out_proj(context_vec)  # optional projection

        return context_vec


class LayerNorm(nn.Module):
    def __init__(self, emb_dim: int):
        super().__init__()
        self.eps = 1e-5
        self.scale = nn.Parameter(torch.ones(emb_dim))
        self.shift = nn.Parameter(torch.zeros(emb_dim))

    @jaxtyping.jaxtyped(typechecker=typechecker)
    def forward(self, x: GPTInnerType) -> GPTInnerType:
        mean = x.mean(dim=-1, keepdim=True)
        var = x.var(dim=-1, keepdim=True, unbiased=False)
        norm_x = (x - mean) / torch.sqrt(var + self.eps)
        return self.scale * norm_x + self.shift


class GELU(nn.Module):
    def __init__(self):
        super().__init__()

    @jaxtyping.jaxtyped(typechecker=typechecker)
    def forward(self, x: GPTInnerType) -> GPTInnerType:
        return 0.5 * x * (1 + torch.tanh(torch.sqrt(torch.tensor(2.0 / torch.pi)) * (x + 0.044715 * torch.pow(x, 3))))


class FeedForward(nn.Module):
    def __init__(self, cfg: GPTModelConfig):
        super().__init__()
        self.layers = nn.Sequential(
            nn.Linear(cfg.emb_dim, 4 * cfg.emb_dim),
            GELU(),
            nn.Linear(4 * cfg.emb_dim, cfg.emb_dim),
            nn.Dropout(cfg.drop_rate),
        )

    @jaxtyping.jaxtyped(typechecker=typechecker)
    def forward(self, x: GPTInnerType) -> GPTInnerType:
        return self.layers(x)


class TransformerBlock(nn.Module):
    def __init__(self, cfg: GPTModelConfig):
        super().__init__()
        self.att = MultiHeadAttention(
            d_in=cfg.emb_dim,
            d_out=cfg.emb_dim,
            ctx_len=cfg.ctx_len,
            n_heads=cfg.n_heads,
            dropout_rate=cfg.drop_rate,
            qkv_bias=cfg.qkv_bias,
        )
        self.ff = FeedForward(cfg)
        self.norm1 = LayerNorm(cfg.emb_dim)
        self.norm2 = LayerNorm(cfg.emb_dim)
        self.drop_resid = nn.Dropout(cfg.drop_rate)

    @jaxtyping.jaxtyped(typechecker=typechecker)
    def forward(self, x: GPTInnerType) -> GPTInnerType:
        # Shortcut connection for attention block
        shortcut = x
        x = self.norm1(x)
        x = self.att(x)  # Shape [batch_size, num_tokens, emb_size]
        x = self.drop_resid(x)
        x = x + shortcut  # Add the original input back

        # Shortcut connection for feed-forward block
        shortcut = x
        x = self.norm2(x)
        x = self.ff(x)
        x = self.drop_resid(x)
        x = x + shortcut  # Add the original input back

        return x


class GPTModel(nn.Module):
    def __init__(self, cfg: GPTModelConfig):
        super().__init__()
        self.config = cfg
        self.tok_emb = nn.Embedding(cfg.vocab_size, cfg.emb_dim)
        self.pos_emb = nn.Embedding(cfg.ctx_len, cfg.emb_dim)
        self.drop_emb = nn.Dropout(cfg.drop_rate)

        self.trf_blocks = nn.Sequential(*[TransformerBlock(cfg) for _ in range(cfg.n_layers)])

        self.final_norm = LayerNorm(cfg.emb_dim)
        self.out_head = nn.Linear(cfg.emb_dim, cfg.vocab_size, bias=False)

    @jaxtyping.jaxtyped(typechecker=typechecker)
    def forward(self, input_vocab_indexes: GPTInputType) -> GPTOutputType:
        _batch_size, num_tokens = input_vocab_indexes.shape
        # batch_size num_tokens -> batch_size num_tokens emb_dim
        tok_embeds = self.tok_emb(input_vocab_indexes)
        # pos_embeds shape: (num_tokens, emb_dim)
        pos_embeds = self.pos_emb(torch.arange(num_tokens, device=input_vocab_indexes.device))
        # pos_embeds is **broadcast** to (batch_size, num_tokens, emb_dim)
        # x: (batch_size, num_tokens, emb_dim)
        x = tok_embeds + pos_embeds  # Shape [batch_size, num_tokens, emb_size]
        x = self.drop_emb(x)
        x = self.trf_blocks(x)
        x = self.final_norm(x)
        logits = self.out_head(x)
        return logits

    @property
    def device(self) -> torch.device:
        return next(self.parameters()).device



import logging
import pathlib
import urllib.request

import tiktoken
import torch
from torch.utils.data import DataLoader, Dataset

logger = logging.getLogger(__name__)


def get_dataset_dir() -> pathlib.Path:
    return pathlib.Path(__file__).parents[1] / "dataset"


class GPTDataset(Dataset):
    def __init__(self, txt: str, tokenizer: tiktoken.Encoding, max_length: int, stride: int):
        """
        :param txt: txt data
        :param tokenizer: tokenizer object
        :param max_length: max length
        :param stride: stride size
        """
        self.tokenizer = tokenizer
        self.input_ids = []
        self.target_ids = []

        # Tokenize the entire text
        token_ids = tokenizer.encode(txt)

        # Use a sliding window to chunk the book into overlapping sequences of max_length
        for i in range(0, len(token_ids) - max_length, stride):
            input_chunk = token_ids[i : i + max_length]
            target_chunk = token_ids[i + 1 : i + max_length + 1]
            self.input_ids.append(torch.tensor(input_chunk))
            self.target_ids.append(torch.tensor(target_chunk))

    def __len__(self):
        return len(self.input_ids)

    def __getitem__(self, idx):
        return self.input_ids[idx], self.target_ids[idx]


class GPTDataloader:
    def __init__(
        self,
        tokenizer: tiktoken.Encoding,
        max_length: int,
        stride: int,
        batch_size: int,
    ):
        self.tokenizer = tokenizer
        self.max_length = max_length
        self.stride = stride
        self.batch_size = batch_size

    def create_dataloader(self, text: str, shuffle=True, drop_last=True) -> DataLoader:
        # Create dataset
        dataset = GPTDataset(text, self.tokenizer, self.max_length, self.stride)

        # Create dataloader
        dataloader = DataLoader(dataset, batch_size=self.batch_size, shuffle=shuffle, drop_last=drop_last)

        return dataloader


def read_simple_text_file() -> str:
    file_name = "the-verdict.txt"
    url = "https://raw.githubusercontent.com/rasbt/LLMs-from-scratch/main/ch02/01_main-chapter-code/the-verdict.txt"

    file_path = get_dataset_dir() / file_name

    # if not exists, download it first
    if not file_path.exists():
        logger.info(f"Downloading {url} to {file_path}")
        with urllib.request.urlopen(url) as response:
            text_data = response.read().decode("utf-8")
        with open(file_path, "w", encoding="utf-8") as file:
            file.write(text_data)

        logger.info(f"Saved {file_path}")
    # open the file
    with open(file_path, "r", encoding="utf-8") as file:
        text_data = file.read()
    return text_data


if __name__ == "__main__":
    text = read_simple_text_file()
    print(len(text))



import jaxtyping
import tiktoken
import torch
from typeguard import typechecked as typechecker


def get_gpt2_tokenizer() -> tiktoken.Encoding:
    tokenizer = tiktoken.get_encoding("gpt2")
    return tokenizer


gpt2_tokenizer = get_gpt2_tokenizer()


@jaxtyping.jaxtyped(typechecker=typechecker)
def text_to_token_ids(
    text: str,
    tokenizer: tiktoken.Encoding = gpt2_tokenizer,
) -> jaxtyping.Int[torch.Tensor, "1 num_tokens"]:
    """
    >>> text_to_token_ids("Hello World", get_gpt2_tokenizer())
    tensor([[15496,  2159]])
    """
    text_id_list = tokenizer.encode(text)
    text_id_tensor = torch.tensor(text_id_list).unsqueeze(0)  # add batch dimension
    return text_id_tensor


@jaxtyping.jaxtyped(typechecker=typechecker)
def token_ids_to_text(
    text_id_tensor: jaxtyping.Int[torch.Tensor, "1 num_tokens"],
    tokenizer: tiktoken.Encoding = gpt2_tokenizer,
) -> str:
    """
    >>> token_ids_to_text(torch.tensor([[15496,  2159]]), get_gpt2_tokenizer())
    'Hello World'
    """
    text_id_list = text_id_tensor.squeeze(0)  # remove batch dimension
    text = tokenizer.decode(text_id_list.tolist())
    return text


if __name__ == "__main__":
    import doctest

    doctest.testmod(verbose=True)



# Copyright 2024 Xiangzhuang Shen
# Copyright (c) Sebastian Raschka under Apache License 2.0 (see LICENSE.txt).
# Source for "Build a Large Language Model From Scratch"
#   - https://www.manning.com/books/build-a-large-language-model-from-scratch
# Code: https://github.com/rasbt/LLMs-from-scratch

from typing import Tuple

import matplotlib.pyplot as plt
import torch
from torch.utils.data import DataLoader

from toyllm.dataset import GPTDataloader
from toyllm.device import current_device
from toyllm.model.config import GPTModelConfig, GPTTrainingConfig
from toyllm.model.generate import TextGenerator
from toyllm.model.gpt import GPTModel
from toyllm.tokenizer import get_gpt2_tokenizer


def get_data_loaders(
    text: str,
    gpt_data_loader: GPTDataloader,
    train_ratio: float = 0.9,
) -> Tuple[DataLoader, DataLoader]:
    # set train/validation split index by train_ratio
    split_idx = int(train_ratio * len(text))
    train_loader = gpt_data_loader.create_dataloader(text=text[:split_idx], drop_last=True, shuffle=True)
    validation_loader = gpt_data_loader.create_dataloader(text=text[:split_idx], drop_last=False, shuffle=False)
    return train_loader, validation_loader


def calc_loss_batch(input_batch, target_batch, model):
    input_batch, target_batch = input_batch.to(current_device), target_batch.to(current_device)
    logits = model(input_batch)
    loss = torch.nn.functional.cross_entropy(logits.flatten(0, 1), target_batch.flatten())
    return loss


def calc_loss_loader(data_loader, model, num_batches=None):
    total_loss = 0.0
    if num_batches is None:
        num_batches = len(data_loader)
    else:
        num_batches = min(num_batches, len(data_loader))
    for i, (input_batch, target_batch) in enumerate(data_loader):
        if i < num_batches:
            loss = calc_loss_batch(input_batch, target_batch, model)
            total_loss += loss.item()
        else:
            break
    return total_loss / num_batches


def evaluate_model(model, train_loader, val_loader, eval_iter):
    model.eval()
    with torch.no_grad():
        train_loss = calc_loss_loader(train_loader, model, num_batches=eval_iter)
        val_loss = calc_loss_loader(val_loader, model, num_batches=eval_iter)
    model.train()
    return train_loss, val_loss


def generate_and_print_sample(model, tokenizer, start_context):
    model.eval()
    text_generate = TextGenerator(model_instance=model, tokenizer=tokenizer)
    generate_text = text_generate.generate(prompt_text=start_context, max_gen_tokens=50, temperature=0.9, top_k=10)
    print(generate_text.replace("\n", " "))  # Compact print format
    model.train()


def train_model_simple(model, train_loader, val_loader, optimizer, num_epochs, eval_freq, eval_iter, start_context):
    # Initialize lists to track losses and tokens seen
    train_losses, val_losses, track_tokens_seen = [], [], []
    tokens_seen = 0
    global_step = -1

    # Main training loop
    for epoch in range(num_epochs):
        model.train()  # Set model to training mode

        for input_batch, target_batch in train_loader:
            optimizer.zero_grad()  # Reset loss gradients from previous epoch
            loss = calc_loss_batch(input_batch, target_batch, model)
            loss.backward()  # Calculate loss gradients
            optimizer.step()  # Update model weights using loss gradients
            tokens_seen += input_batch.numel()
            global_step += 1

            # Optional evaluation step
            if global_step % eval_freq == 0:
                train_loss, val_loss = evaluate_model(model, train_loader, val_loader, eval_iter)
                train_losses.append(train_loss)
                val_losses.append(val_loss)
                track_tokens_seen.append(tokens_seen)
                print(
                    f"Ep {epoch+1} (Step {global_step:06d}): " f"Train loss {train_loss:.3f}, Val loss {val_loss:.3f}"
                )

        # Print a sample text after each epoch
        generate_and_print_sample(model, train_loader.dataset.tokenizer, start_context)

    return train_losses, val_losses, track_tokens_seen


def plot_losses(epochs_seen, tokens_seen, train_losses, val_losses):
    fig, ax1 = plt.subplots()

    # Plot training and validation loss against epochs
    ax1.plot(epochs_seen, train_losses, label="Training loss")
    ax1.plot(epochs_seen, val_losses, linestyle="-.", label="Validation loss")
    ax1.set_xlabel("Epochs")
    ax1.set_ylabel("Loss")
    ax1.legend(loc="upper right")

    # Create a second x-axis for tokens seen
    ax2 = ax1.twiny()  # Create a second x-axis that shares the same y-axis
    ax2.plot(tokens_seen, train_losses, alpha=0)  # Invisible plot for aligning ticks
    ax2.set_xlabel("Tokens seen")

    fig.tight_layout()  # Adjust layout to make room
    plt.show()


def main(
    text: str,
    gpt_config: GPTModelConfig,
    training_config: GPTTrainingConfig,
):
    torch.manual_seed(123)
    # Initialize model
    model = GPTModel(gpt_config)
    model.to(current_device)
    optimizer = torch.optim.AdamW(
        model.parameters(), lr=training_config.learning_rate, weight_decay=training_config.weight_decay
    )
    tokenizer = get_gpt2_tokenizer()

    gpt_data_loader = GPTDataloader(
        tokenizer=tokenizer,
        max_length=gpt_config.ctx_len,
        stride=gpt_config.ctx_len,
        batch_size=training_config.batch_size,
    )
    train_loader, val_loader = get_data_loaders(text, gpt_data_loader=gpt_data_loader, train_ratio=0.9)

    # Train model
    train_losses, val_losses, tokens_seen = train_model_simple(
        model,
        train_loader,
        val_loader,
        optimizer,
        num_epochs=training_config.num_epochs,
        eval_freq=5,
        eval_iter=1,
        start_context="Every effort moves you",
    )

    return train_losses, val_losses, tokens_seen, model


if __name__ == "__main__":
    from toyllm.dataset import read_simple_text_file
    from toyllm.model.config import GPTTrainingConfig, gpt_config_124_m

    training_config = GPTTrainingConfig(learning_rate=5e-4, num_epochs=40, batch_size=2, weight_decay=0.1)

    text = read_simple_text_file()
    # Initiate training
    train_losses, val_losses, tokens_seen, model = main(text, gpt_config_124_m, training_config)

    # After training

    # Plot results
    epochs_tensor = torch.linspace(0, training_config.num_epochs, len(train_losses))
    plot_losses(epochs_tensor, tokens_seen, train_losses, val_losses)
    plt.savefig("loss.pdf")

    # Save and load model
    # torch.save(model.state_dict(), "model.pth")
    # model = GPTModel(gpt_config_124_m)
    # model.load_state_dict(torch.load("model.pth"))



from dataclasses import dataclass


@dataclass
class GPTModelConfig:
    """
    GPT Model Architecture Config

    :param vocab_size: vocabulary size
    :param ctx_len: length of context/block
    :param emb_dim: embedding size
    :param n_heads: number of attention heads
    :param n_layers: number of transformer layers
    :param drop_rate: dropout rate
    :param qkv_bias: query key value bias terms
    """

    vocab_size: int
    ctx_len: int
    emb_dim: int
    n_heads: int
    n_layers: int
    drop_rate: float
    qkv_bias: bool


gpt_config_124_m = GPTModelConfig(
    vocab_size=50257,
    ctx_len=1024,
    emb_dim=768,
    n_heads=12,
    n_layers=12,
    drop_rate=0.1,
    qkv_bias=False,
)


@dataclass
class GPTTrainingConfig:
    """
    GPT training config: hyperparameters for GPT model training
    """

    learning_rate: float = 5e-4
    num_epochs: int = 10
    batch_size: int = 2
    weight_decay: float = 0.1
